"use strict";
/**
 *  Copyright (c) 2019, 2024 Texas Instruments Incorporated
 *  All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *      Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *  notice, this list of conditions and the following disclaimer in the
 *  documentation and/or other materials provided with the distribution.
 *      Neither the name of Texas Instruments Incorporated nor the names of
 *  its contributors may be used to endorse or promote products derived
 *  from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 *  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  Original Author:
 *      Patrick Chuong, Texas Instruments, Inc.
 */
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
/* eslint-disable no-console */
/* eslint-disable @typescript-eslint/no-explicit-any */
const optimist_1 = __importDefault(require("optimist"));
const path_1 = __importDefault(require("path"));
const os_1 = __importStar(require("os"));
const fs_extra_1 = __importDefault(require("fs-extra"));
const child_process_1 = __importDefault(require("child_process"));
const TargetConfig_1 = require("./TargetConfig");
const json5_1 = __importDefault(require("json5"));
const express_1 = __importDefault(require("express"));
// @ts-ignore
const open_1 = __importDefault(require("open"));
/**
 * Helper for converting command line arguments to absolute path, including support for home tilda (~) prefix.
 * The tilda (~) is not part of the OS, and is a feature of command shell(s) only, so it has to be
 * removed from the path.  Path.resolve() or path.join() does not handle this.
 */
function convertToAbsolutePath(relPath, absRoot = __dirname) {
    // convert ~ to homedir
    relPath = relPath.replace(/^~($|\/|\\)/, `${os_1.homedir()}$1`);
    if (!path_1.default.isAbsolute(relPath)) {
        relPath = path_1.default.join(absRoot, relPath);
    }
    return relPath;
}
/**
 * supported launcher.json parameters
 */
const PARAM_BROWSER = 'browser';
const PARAM_PORT = 'port';
const PARAM_SCRIPT = 'script';
const PARAM_LOGFILE = 'logfile';
const browser = optimist_1.default.argv[PARAM_BROWSER] ?? '';
const localhost = '127.0.0.1';
const isNW = browser.indexOf('nw') > 0;
const instanceOfAddress = (arg) => {
    return !!arg.port;
};
/**
 * Get the app profile path
 */
const getProfilePath = (app) => {
    const getWin = () => path_1.default.join(os_1.homedir(), 'AppData', 'Local');
    const getLinux = () => path_1.default.join(os_1.homedir(), '.config');
    const getOSX = () => path_1.default.join(os_1.homedir(), 'Library', 'Application Support');
    const getDefault = () => os_1.platform().startsWith('win') ? getWin() : getLinux();
    let appDataPath = process.env['APPDATA'];
    if (!appDataPath) {
        switch (os_1.platform()) {
            case 'win32':
                appDataPath = getWin();
                break;
            case 'linux':
                appDataPath = getLinux();
                break;
            case 'darwin':
                appDataPath = getOSX();
                break;
            default:
                appDataPath = getDefault();
                break;
        }
    }
    return path_1.default.join(appDataPath, app);
};
/**
 * Returns the cloudagent platform.
 */
const getAgentPlatform = () => {
    switch (os_1.platform()) {
        case 'win32': return 'win';
        case 'linux': return 'linux';
        case 'darwin': return 'osx';
        default: return 'unsupported';
    }
};
/**
 * Splash screen
 */
if (isNW && !/^darwin/.test(process.platform)) {
    (async () => {
        const cleanup = async () => {
            const profilePath = getProfilePath('GUI Composer Splash' /* app/splash/package.json name property */);
            await fs_extra_1.default.remove(profilePath);
        };
        const splashProcess = await open_1.default('', { app: [convertToAbsolutePath(browser), path_1.default.join(__dirname, 'splash')] });
        splashProcess.on('exit', cleanup.bind(null));
        splashProcess.on('SIGINT', cleanup.bind(null));
        splashProcess.on('SIGUSR1', cleanup.bind(null));
        splashProcess.on('SIGUSR2', cleanup.bind(null));
        splashProcess.on('uncaughtException', cleanup.bind(null));
    })();
}
class ApplicationServer {
    constructor() {
        this.app = express_1.default();
        this.router = express_1.default.Router();
        this.TICloudAgentPublicDir = path_1.default.join(__dirname, '../runtime/ticloudagent/server/public');
        this.targetConfig = new TargetConfig_1.TargetConfig(path_1.default.resolve(__dirname, '../runtime'));
    }
    async initializeRoutes(router) {
        /* respond headers */
        router.use((req, res, next) => {
            /* enable shared array buffer for scripting */
            res.setHeader('Cross-Origin-Opener-Policy', 'same-origin');
            res.setHeader('Cross-Origin-Embedder-Policy', 'require-corp');
            next();
        });
        router.get('**/version.xml', async (req, res) => {
            const content = await fs_extra_1.default.readFile(path_1.default.join(__dirname, '../version.xml'), 'utf-8');
            res.header('Content-Type', 'text/xml');
            res.send(content);
        });
        /*
         * server config route
         */
        router.get('**/server-config.js', (req, res) => {
            const serverConfigObj = {
                isOnline: false
            };
            let serverConfig = 'window.gc = window.gc || {};\n';
            serverConfig += `window.gc.serverConfig = ${JSON.stringify(serverConfigObj)};`;
            res.header('Content-Type', 'text/javascript');
            res.send(serverConfig);
        });
        /*
         * handles ticloudagent routes
         */
        router.get('/ticloudagent/agent_config.json', async (req, res) => {
            try {
                const agentOutput = await this.startCloudAgentHostApp();
                res.header('Content-Type', 'application/json');
                res.send(agentOutput);
            }
            catch (e) {
                res.sendStatus(404);
            }
        });
        router.get(['/ticloudagent/getConfigInfo', '/ticloudagent/getConfigInfo/'], async (req, res) => {
            try {
                const result = await this.targetConfig.getConfigInfo(getAgentPlatform());
                res.set('Content-Disposition', 'attachment;filename=target_setup.json');
                res.send(result);
            }
            catch (e) {
                res.sendStatus(404);
            }
        });
        router.get('/ticloudagent/getConfig/**', async (req, res) => {
            try {
                const [platform, connectionId, devicePath, deviceId] = req.path.replace('/ticloudagent/getConfig/', '').split('/');
                const xmlPath = devicePath === 'boards' ? 'boards' : 'devices';
                const xmlFile = devicePath === 'boards' || devicePath === 'devices' ? deviceId : devicePath;
                const result = await this.targetConfig.getConfig(platform, connectionId, xmlPath, xmlFile);
                res.set('Content-Disposition', 'attachment;filename=target.ccxml');
                res.send(result);
            }
            catch (e) {
                res.writeHead(500, {
                    'content-type': 'text/html; charset=utf8',
                });
                res.end('' + e);
            }
        });
        router.get('/ticloudagent/getFile/**', async (req, res) => {
            try {
                const segments = req.path.replace('/ticloudagent/getFile/', '').split('/');
                const platform = segments.splice(0, 1)[0];
                const version = segments.pop();
                const filepath = segments.join('/');
                const filename = segments[segments.length - 1];
                const result = await this.targetConfig.getFile(platform, filepath, version);
                res.set('Content-Disposition', `attachment;filename=${filename}`);
                res.send(result);
            }
            catch (e) {
                res.sendStatus(404);
            }
        });
        router.get('/ticloudagent/**', async (req, res) => {
            try {
                const reqPath = req.url.substring('/ticloudagent'.length + 1);
                const content = await fs_extra_1.default.readFile(path_1.default.join(this.TICloudAgentPublicDir, reqPath), 'utf-8');
                res.header('Content-Type', 'application/javascript');
                res.send(content);
            }
            catch (e) {
                res.sendStatus(404);
            }
        });
        /* appscript route */
        if (optimist_1.default.argv[PARAM_SCRIPT]) {
            router.get('/appscript', async (req, res) => {
                try {
                    const filepath = convertToAbsolutePath(optimist_1.default.argv[PARAM_SCRIPT].trim());
                    const content = await fs_extra_1.default.readFile(filepath, 'utf-8');
                    res.header('Content-Type', 'text/plain');
                    res.send(content);
                }
                catch (e) {
                    res.sendStatus(404);
                }
            });
        }
        /* log file route */
        if (optimist_1.default.argv[PARAM_LOGFILE]) {
            router.get('/appscript', async (req, res) => {
                const filepath = convertToAbsolutePath(optimist_1.default.argv[PARAM_LOGFILE].trim());
                res.header('Content-Type', 'text/plain');
                res.send(filepath);
            });
        }
        /*
         * handles /sysconfig/ routes
         */
        const sdkPaths = optimist_1.default.argv['sdkPaths'];
        const sdkRoots = (sdkPaths ? sdkPaths.split(',') : []).map(sdkPath => convertToAbsolutePath(sdkPath.trim()));
        router.get(['/sysconfig/fetchAvailableProducts'], async (req, res) => {
            try {
                const results = [];
                for (let i = 0; i < sdkRoots.length; i++) {
                    const sdkRoot = sdkRoots[i];
                    if (!fs_extra_1.default.existsSync(sdkRoot)) {
                        continue;
                    }
                    const files = await fs_extra_1.default.readdir(sdkRoot);
                    for (const name of files) {
                        const fileName = fs_extra_1.default.existsSync(path_1.default.join(sdkRoot, name, '.metadata', 'product.json')) ? 'product.json' : 'sdk.json';
                        const pathSegments = [`/availableProduct${i || ''}`, name, '.metadata', fileName];
                        const location = path_1.default.join(sdkRoot, ...pathSegments.slice(1));
                        try {
                            const { name, version, displayName } = json5_1.default.parse(await fs_extra_1.default.readFile(location, 'utf-8') || '');
                            if (name && version) {
                                results.push({ name, version, displayName: displayName ?? name, path: pathSegments.join('/'), location });
                            }
                        }
                        catch (e) {
                            // ignore errors reading individual product.json files.
                        }
                    }
                }
                res.header('Content-Type', 'application/json');
                res.send(JSON.stringify(results));
            }
            catch (e) {
                res.writeHead(500, {
                    'content-type': 'text/html; charset=utf8',
                });
                res.end('' + e);
            }
        });
        /*
         * add /sysconfig/ prefix to availableProducts routes, which can occur when images are loaded from the sdk
         * within markdown module descriptions.
         */
        router.get('/availableProduct*', (req, res, next) => {
            req.url = '/sysconfig' + req.url;
            next();
        });
        sdkRoots.forEach((sdkRoot, i) => {
            router.use(`/sysconfig/availableProduct${i || ''}/`, express_1.default.static(sdkRoot));
        });
        let sysconfigPath = optimist_1.default.argv['sysconfig']?.trim();
        if (sysconfigPath) {
            sysconfigPath = convertToAbsolutePath(sysconfigPath);
            if (await fs_extra_1.default.pathExists(path_1.default.join(sysconfigPath, 'out', 'dist'))) {
                sysconfigPath = path_1.default.join(sysconfigPath, 'out', 'dist');
            }
            else if (await fs_extra_1.default.pathExists(path_1.default.join(sysconfigPath, 'dist'))) {
                sysconfigPath = path_1.default.join(sysconfigPath, 'dist');
            }
        }
        router.get('/sysconfig/getPath', async (req, res) => {
            try {
                if (!sysconfigPath) {
                    throw new Error('Missing sysconfig argument on command line.');
                }
                res.header('Content-Type', 'text/plain');
                res.send(sysconfigPath);
            }
            catch (e) {
                res.sendStatus(404);
            }
        });
        if (!sysconfigPath) {
            router.use('/sysconfig/**', (req, res) => res.sendStatus(404));
        }
        else {
            router.use('/sysconfig/', express_1.default.static(sysconfigPath));
        }
        /* api route */
        router.get('/api/shutdown', (req, res) => {
            console.log('Server shutting received...');
            res.header('Content-Type', 'text/plain');
            res.send('OK');
            if (!isNW) { /* shutdown for non-nodewebkit browser */
                this.shutdown(0);
            }
        });
        /* default route */
        router.use(express_1.default.static(__dirname, { index: ['index.html', 'index.htm'], extensions: ['js'] }));
    }
    shutdown(code) {
        console.debug(`GUI Composer Server terminated, exit code = ${code}`);
        process.exit(code);
    }
    async startCloudAgentHostApp() {
        if (!this.agentHostAppPromise) {
            const ext = os_1.platform() === 'win32' ? 'bat' : 'sh';
            const cwd = path_1.default.join(__dirname, '../runtime/TICloudAgentHostApp');
            const executable = path_1.default.join(`${cwd}/ticloudagent.${ext}`);
            this.agentHostAppPromise = new Promise((resolve, reject) => {
                const app = child_process_1.default.spawn(executable, ['not_chrome'], { cwd: cwd });
                app.stdout.on('data', data => resolve(data.toString()));
                app.stderr.on('data', data => {
                    // cloudagent returns an error message 'Debugger attached.' when debugging in VSCode,
                    // however it is still functional without any issues.
                    const message = data.toString();
                    if (message.trim() !== 'Debugger attached.') {
                        reject(message);
                    }
                });
            }).then(data => {
                const _data = JSON.parse(data);
                _data.offline = true;
                _data.agentPort = _data.port;
                delete _data.port;
                console.log(`CloudAgent: ${JSON.stringify(_data)}`);
                return _data;
            });
        }
        return await this.agentHostAppPromise;
    }
    /**
     * Run the GC App server.
     */
    async listen() {
        let port = optimist_1.default.argv[PARAM_PORT] || 0;
        await new Promise((resolve) => {
            const server = this.app.listen(port, '127.0.0.1', async () => {
                /* sanity check for the listening port */
                const address = server.address();
                if (instanceOfAddress(address)) {
                    port = address.port;
                    console.log(`GUI Composer Application Server listening at port: ${port}`);
                    resolve();
                }
                else {
                    console.error(`... failed to start server at port: ${port}`);
                    process.exit(1);
                }
            });
        });
        await this.initializeRoutes(this.router);
        this.app.use('/', this.router);
        if (!browser) {
            return; // headless, no browser.  Used for running UI selenium tests.
        }
        const url = `http://${localhost}:${port}`;
        console.log(`Launching browser ${browser} @ ${url}`);
        if (isNW) {
            /* read the package.json and create a temp folder using the app name */
            const packageJson = await fs_extra_1.default.readJson(path_1.default.join(__dirname, 'package.json'));
            let projectJson = {};
            try {
                projectJson = await fs_extra_1.default.readJson(path_1.default.join(__dirname, 'project.json'));
            }
            catch (e) {
                console.log('Warning: missing project.json file');
            }
            const appName = (projectJson.projectName || packageJson.name).replace(/\s+/g, '_');
            const tmpDir = await fs_extra_1.default.mkdtemp(path_1.default.join(os_1.default.tmpdir(), appName + '_'));
            /* set the url include the port number */
            packageJson.main = url;
            /* copy the icon file - nw.exe require it to be relative to the package.json */
            if (packageJson.window && projectJson.windowIcon) {
                const legacyWindowIconPath = path_1.default.join(__dirname, 'images', projectJson.windowIcon); // legacy project with images folder
                const srcIconPath = fs_extra_1.default.existsSync(legacyWindowIconPath) ? legacyWindowIconPath : path_1.default.join(__dirname, 'assets', projectJson.windowIcon);
                const destIconPath = path_1.default.join(tmpDir, 'assets', projectJson.windowIcon);
                await fs_extra_1.default.mkdirp(path_1.default.dirname(destIconPath));
                fs_extra_1.default.createReadStream(srcIconPath).pipe(fs_extra_1.default.createWriteStream(destIconPath));
                packageJson.window.icon = 'assets/' + projectJson.windowIcon;
            }
            /* write the package.json to the temp folder */
            await fs_extra_1.default.writeFile(path_1.default.join(tmpDir, 'package.json'), JSON.stringify(packageJson, undefined, 2));
            /* clean up temp folder and terminate process */
            const cleanup = async () => {
                await fs_extra_1.default.remove(tmpDir);
                await fs_extra_1.default.remove(path_1.default.join(getProfilePath(packageJson.name), os_1.platform() === 'win32' ? 'User Data' : '', 'Default', 'Web Data'));
                this.shutdown(0);
            };
            process.on('exit', cleanup.bind(null));
            process.on('SIGINT', cleanup.bind(null));
            process.on('SIGUSR1', cleanup.bind(null));
            process.on('SIGUSR2', cleanup.bind(null));
            process.on('uncaughtException', cleanup.bind(null));
            /* launch nw */
            await open_1.default('', { app: [convertToAbsolutePath(browser), tmpDir], wait: true });
            /* when nw or browser process exit, kill ourself */
            cleanup();
        }
        else {
            /* launch browser */
            await open_1.default(url, { app: browser, wait: true });
        }
    }
}
new ApplicationServer().listen();
//# sourceMappingURL=ApplicationServer.js.map